import secrets
import logging

from .runpipe import RunPipe
from ..pipelines.customimputation import imps
from ..pipelines.customclf import clfs
from ..pipelines.customrsp import rsps
from ..pipelines.customhbd import hbds
from ..pipelines.customautoml import automls
from ..components.hybrids import RunAutoSmote
from ..processing.utils import Result
from ..processing.selectiontree import BinaryTree
from ..components.model_filters import ModelFiltering


class AutoImblearn:
    """ The core class that defines how to search the optimal pipeline given a dataset
    Parameters
    ----------
    run_pipe : RunPipe class
        The helper class that has two functions:
            1) handles loading, splitting the data
            2) run the pipelines and get the results

    metric : The evaluation metric defined by user to use during training and final evaluation

    Attributes
    ----------
    resamplers : The list of re-samplers available to choose from

    classifiers : The list of classifiers available to choose from

    hybrids : The list of hybrid re-sampler and classifier methods available to choose from

    imputers : The list of imputers available to choose from

    automls : The list of AutoMLs available to choose from

    run_pipe : RunPipe class

    metric : Evaluation metric
    """
    def __init__(self, run_pipe: RunPipe, metric):
        self.resamplers = list(rsps.keys())
        self.classifiers = list(clfs.keys())
        self.hybrids = list(hbds.keys())
        self.automls = list(automls.keys())
        self.imputers = imps
        self.run_pipe = run_pipe
        self.metric = metric
        self.local_best_pipes = BinaryTree()

    def _train_automl(self, pipeline):
        """ Train the selected automl model """
        if len(pipeline) != 1:
            raise ValueError("Pipeline {} length is not correct, not an automl model")
        # TODO test this function
        result = self.run_pipe.fit_automl(pipeline)
        return result

    def _train_hybird(self, pipeline, args=None, train_ratio=1.0):
        """ Train the pipeline with hybrid resampling method """
        if len(pipeline) != 2:
            raise ValueError("Pipeline {} length is not correct, not a hybrid method pipeline")
        result = self.run_pipe.fit_hybrid(pipeline)
        return result

    def _train_regular(self, pipeline):
        """ Train the pipeline with regular (imputer, resampler, classifier) pipeline """
        if len(pipeline) != 3:
            raise ValueError("Pipeline {} length is not correct, not a regular method pipeline")
        result = self.run_pipe.fit(pipeline)
        return result

    def _init_result_tree(self, result_tree: BinaryTree):
        imp = secrets.choice(self.imputers)
        result_tree.insert(imp, -1, "imp", "root")

        rsp = secrets.choice(self.resamplers)
        result_tree.insert(rsp, -1, "rsp", "imp")

        clf = secrets.choice(self.classifiers)
        result_tree.insert(clf, -1, "clf", "rsp")

        hbd = secrets.choice(self.hybrids)
        result_tree.insert(hbd, -1, "hbd", "imp")

        automl = secrets.choice(self.automls)
        result_tree.insert(automl, -1, "automl", "root")

        # result_tree.print_tree()

    def _compute_current_pipe(self, result_tree: BinaryTree):
        """ Select a path from 3 different pipe types """
        pool = ["automl", "hbd", "clf"]
        pipe_type = secrets.choice(pool)
        pipe = result_tree.build_pipe(pipe_type)
        result_tree.replace(pipe_type, 0, pipe[-1])


        self.model_filtering(topn=3)

    def model_filtering(self, topn=3):
        # Get dataset description
        dp = self.run_pipe.dataloader.get_data_description(self.run_pipe.args.dataset)

        # Get model description TODO

        # Filter models
        model_filter = ModelFiltering(dp, self.run_pipe.dataloader.get_data_folder())
        # TODO make this work
        # filtered_models = model_filter.get_topn()
        # print(filtered_models)

        imputers = model_filter.get_topn("imputer")
        self.imputers = [imputer for imputer in imputers if imputer in self.imputers]

        resamplers = model_filter.get_topn("resampler")
        self.resamplers = [resampler for resampler in resamplers if resampler in self.resamplers]

        classifiers = model_filter.get_topn("classifier")
        self.classifiers = [classifier for classifier in classifiers if classifier in self.classifiers]


    def exhaustive_search(self, checked=None, train_ratio=1.0):
        saver = Result(train_ratio, self.metric)
        saver.load_saved_result()

        for imp in self.imputers:
            for resampler in self.resamplers:
                for classifier in self.classifiers:
                    pipe = [imp, resampler, classifier]
                    print(pipe)
                    if is_checked(pipe, checked):
                        tmp = checked[imp][resampler][classifier]
                    else:
                        if saver.is_in(pipe):
                            tmp = saver.get(pipe)
                        else:
                            try:
                                if resampler == "autosmote":
                                    run_autosmote = RunAutoSmote()
                                    tmp = run_autosmote.fit(clf=classifier, imp=imp, metric=self.metric, train_ratio=train_ratio)
                                else:
                                    tmp = self.run_pipe.fit(pipe)
                            except Exception as e:
                                raise e
                                tmp = 0

                            saver.append(pipe, tmp)
                        checked[imp][resampler][classifier] = tmp
                    print("Current pipe: {}, result: {}".format(pipe, tmp))


    def find_best(self, checked=None, train_ratio=1.0):
        # saver = Result(train_ratio, self.metric, self.run_pipe.args.dataset)
        saver = self.run_pipe.saver
        saver.load_saved_result()

        result_tree = BinaryTree()
        self._init_result_tree(result_tree)
        self._compute_current_pipe(result_tree)
        current_pipe = result_tree.best_pipe()
        # print(current_pipe)
        # result_tree.print_tree()

        counter = 0
        # TODO test saving result , 1) add auto-sklearn to automls (using flask rest API)
        #                                from flask import Flask, jsonify
        best_pipe = []
        best_score = 0
        final_result = set([])

        def update_best_pipe(tmp_pipe, result):
            nonlocal best_pipe, best_score, current_pipe
            if result > best_score:
                best_pipe = list(tmp_pipe)
                best_score = result
            if result_tree.update_pipe(tmp_pipe, result):
                current_pipe = result_tree.best_pipe()
            logging.info(f"This is the best result so far: {best_score} {best_pipe}, This is the current result: {result}, {tmp_pipe}")

        def train_and_update(tmp_pipe):
            nonlocal counter
            if saver.is_in(tmp_pipe):
                result = saver.get(tmp_pipe)
            else:
                if len(tmp_pipe) == 2:
                    result = self._train_hybird(tmp_pipe)
                elif len(tmp_pipe) == 3:
                    result = self._train_regular(tmp_pipe)
                else:
                    result = self._train_automl(tmp_pipe)
                saver.append(tmp_pipe, result)
                counter += 1
            update_best_pipe(tmp_pipe, result)
            return result


        while True:
            # Brute force method
            # Step 1: Choose imputater or other automl
            # for model in self.imputers + self.automls:
            for model in self.imputers:
                if model in self.imputers:
                    # Regular or Hybrid method
                    tmp_pipe = result_tree.sub_best_pipe(node_type="imp")
                    if len(tmp_pipe) == 2:
                        tmp_pipe = [model] + tmp_pipe[1:]

                        result = train_and_update(tmp_pipe)

                    elif len(tmp_pipe) == 3:
                        tmp_pipe = [model] + tmp_pipe[1:]
                        result = train_and_update(tmp_pipe)
                    else:
                        raise ValueError(
                            "Pipeline length of {} is not compatible with AutoImblearn".format(current_pipe))

                else:
                    # AutoML method
                    tmp_pipe = [model]
                    result = train_and_update(tmp_pipe)


                # print("This is the best result so far: ", best_score, best_pipe, "This is the current result: ", result, tmp_pipe)
                # print(result, best_score)
                if result > best_score:
                    best_pipe = list(tmp_pipe)
                    best_score = result

            if len(current_pipe) != 1:
                # Step 2: Choose resampler
                for model in self.resamplers + self.hybrids:
                    if model in self.resamplers:
                        sub_pipe = result_tree.sub_best_pipe("clf")
                        tmp_pipe = [current_pipe[0], model] + sub_pipe

                        result = train_and_update(tmp_pipe)
                    else:
                        # Hybrid method
                        tmp_pipe = [current_pipe[0], model]
                        result = train_and_update(tmp_pipe)

                    # print("This is the best result so far: ", best_score, best_pipe, "This is the current result: ", result, tmp_pipe)
                    if result > best_score:
                        best_pipe = list(tmp_pipe)
                        best_score = result

            if len(current_pipe) == 3:
                # Step 3: Choose classifier
                for model in self.classifiers:
                    tmp_pipe = current_pipe[:2] + [model]
                    result = train_and_update(tmp_pipe)

                    # print("This is the best result so far: ", best_score, best_pipe, "This is the current result: ", result, tmp_pipe)
                    if result > best_score:
                        best_pipe = list(tmp_pipe)
                        best_score = result

            if set(best_pipe) == set(final_result):
                break
            else:
                final_result = list(best_pipe)

        best_pipe = result_tree.best_pipe()
        result_tree.print_tree()
        return best_pipe, counter, best_score

    def run_best(self, pipeline=None):
        # Re-run the best pipeline found with 100% of data
        # saver = Result(1.0, self.metric, self.run_pipe.args.dataset)
        saver = self.run_pipe.saver
        saver.load_saved_result()
        if saver.is_in(pipeline):
            result = saver.get(pipeline)
        else:
            if len(pipeline) == 1:
                result = self._train_automl(pipeline)
            elif len(pipeline) == 2:
                result = self._train_hybird(pipeline, args, 1.0)
            elif len(pipeline) == 3:
                result = self._train_regular(pipeline)
            else:
                raise Exception("Pipeline {} is not in the correct length".format(pipeline))

        return result

    def count_pipe(self, pipeline=None):
        # Find the optimal and count how many pipelines to check
        counters = []
        for _ in range(100):
            checked = []
            final, count, best_score = self.find_best(checked)
            while final != set(pipeline):
                final, count, best_score = self.find_best(checked)
            counters.append(count)
        return counters


if __name__ == "__main__":
    class Args:
        def __init__(self):
            self.train_ratio=1.0
            self.n_splits = 10
            self.repeat = 0
            self.dataset = "test"
            self.metric = "auroc"
    args = Args()
    run_pipe = RunPipe(args)
    autoimb = AutoImblearn(run_pipe, metric=args.metric)
    tmp_tree = BinaryTree()
    autoimb._init_result_tree(tmp_tree)
    tmp_tree.print_tree()

    autoimb._compute_current_pipe(tmp_tree)

    tmp_pipe = tmp_tree.best_pipe()
    # print(tmp_pipe)
    # tmp_tree.print_tree()
    result = tmp_tree.build_pipe("clf")
    print(result)

    tmp_tree.update_pipe(["ii", "smote", "mlp"], 2)
    tmp_tree.update_pipe(["ii", "under", "svm"], 1)
    tmp_tree.print_tree()




import requests
import pandas as pd
from .base_model_client import BaseDockerModelClient


class BaseEstimator(BaseDockerModelClient):
    """Abstract base class for sklearn-like estimators/classifiers."""

    def payload(self):
        pass

    def predict(self, X):
        """Predict with data X"""
        try:
            self.ensure_container_running()
            payload = {
                # "data": self.(X).to_dict(orient="records")
            }
            response = requests.post(f"{self.api_url}/predict", json=payload)
            response.raise_for_status()
            result = response.json().get("balanced_data", [])
            return pd.DataFrame(result)
        finally:
            self.stop_container()



import logging

import numpy as np
from sklearn import svm
from sklearn.linear_model import LogisticRegression
from sklearn.neural_network import MLPClassifier
from sklearn.ensemble import RandomForestClassifier, AdaBoostClassifier, ExtraTreesClassifier
from sklearn.metrics import precision_recall_fscore_support, roc_auc_score, classification_report, \
    average_precision_score

from AutoImblearn.components.api import BaseEstimatorAPI

clfs = {
    "lr": LogisticRegression(random_state=42, max_iter=500, solver='lbfgs'),
    "mlp": MLPClassifier(alpha=1, random_state=42, max_iter=1000),
    "ada": AdaBoostClassifier(random_state=42),
    "svm": svm.SVC(random_state=42, probability=True, kernel='linear'),
    # "SVM": svm.SVC(random_state=42, probability=True, kernel='poly'),
    # "rf": RandomForestClassifier(random_state=42),
    # "ensemble": XGBClassifier(learning_rate=1.0, max_depth=10, min_child_weight=15, n_estimators=100, n_jobs=1, subsample=0.8, verbosity=0),
    # "bst": GradientBoostingClassifier(random_state=42),
}

hyperparameter_search_space = {
    "lr": {
        "penalty": {
            "type": "categorical",
            "choices": ["l2", "l1", "elasticnet", "none"],
            "default": "l2"
        },
        "C": {
            "type": "float",
            "min": 0.01,
            "max": 10.0,
            "default": 1.0,
            "log_scale": True
        },
        "solver": {
            "type": "categorical",
            "choices": ["lbfgs", "liblinear", "saga", "newton-cg", "sag"],
            "default": "lbfgs"
        },
        "max_iter": {
            "type": "int",
            "min": 100,
            "max": 1000,
            "default": 500
        }
    },
    "mlp": {
        "hidden_layer_sizes": {
            "type": "categorical",
            "choices": [(50,), (100,), (100, 50), (100, 100)],
            "default": (100,)
        },
        "activation": {
            "type": "categorical",
            "choices": ["relu", "tanh", "logistic"],
            "default": "relu"
        },
        "solver": {
            "type": "categorical",
            "choices": ["adam", "sgd", "lbfgs"],
            "default": "adam"
        },
        "alpha": {
            "type": "float",
            "min": 1e-5,
            "max": 1.0,
            "default": 1.0,
            "log_scale": True
        },
        "learning_rate": {
            "type": "categorical",
            "choices": ["constant", "invscaling", "adaptive"],
            "default": "constant"
        },
        "max_iter": {
            "type": "int",
            "min": 200,
            "max": 2000,
            "default": 1000
        }
    },
    "ada": {
        "n_estimators": {
            "type": "int",
            "min": 10,
            "max": 1000,
            "default": 50
        },
        "learning_rate": {
            "type": "float",
            "min": 0.01,
            "max": 2.0,
            "default": 1.0,
            "log_scale": True
        },
        "algorithm": {
            "type": "categorical",
            "choices": ["SAMME", "SAMME.R"],
            "default": "SAMME.R"
        }
    },
    "svm": {
        "C": {
            "type": "float",
            "min": 0.01,
            "max": 10.0,
            "default": 1.0,
            "log_scale": True
        },
        "kernel": {
            "type": "categorical",
            "choices": ["linear", "poly", "rbf", "sigmoid"],
            "default": "linear"
        },
        "degree": {
            "type": "int",
            "min": 2,
            "max": 5,
            "default": 3  # used in poly kernel
        },
        "gamma": {
            "type": "categorical",
            "choices": ["scale", "auto"],
            "default": "scale"
        },
        "coef0": {
            "type": "float",
            "min": 0.0,
            "max": 1.0,
            "default": 0.0
        },
        "probability": {
            "type": "categorical",
            "choices": [True, False],
            "default": True
        }
    },
    "ensemble": {
        "learning_rate": {
            "type": "float",
            "min": 0.01,
            "max": 1.0,
            "default": 1.0,
            "log_scale": True
        },
        "max_depth": {
            "type": "int",
            "min": 1,
            "max": 20,
            "default": 10
        },
        "min_child_weight": {
            "type": "int",
            "min": 1,
            "max": 30,
            "default": 15
        },
        "n_estimators": {
            "type": "int",
            "min": 10,
            "max": 1000,
            "default": 100
        },
        "subsample": {
            "type": "float",
            "min": 0.5,
            "max": 1.0,
            "default": 0.8
        },
        "colsample_bytree": {
            "type": "float",
            "min": 0.5,
            "max": 1.0,
            "default": 1.0
        },
        "gamma": {
            "type": "float",
            "min": 0.0,
            "max": 5.0,
            "default": 0.0
        },
        "reg_alpha": {
            "type": "float",
            "min": 0.0,
            "max": 5.0,
            "default": 0.0
        },
        "reg_lambda": {
            "type": "float",
            "min": 0.0,
            "max": 5.0,
            "default": 1.0
        },
        "n_jobs": {
            "type": "int",
            "min": 1,
            "max": 16,
            "default": 1
        },
        "verbosity": {
            "type": "categorical",
            "choices": [0, 1, 2, 3],
            "default": 0
        }
    }
}

class RunSklearnClassifierAPI(BaseEstimatorAPI):

    def __init__(self, import_name):
        super().__init__(import_name)
        self.clf = None

    def get_hyperparameter_search_space(self):
        clf = self.params.model
        return hyperparameter_search_space[clf]

    def fit(self, args, X_train, y_train, X_test, y_test):
        clf = self.params.model
        params = self.dict_to_namespace()

        try:
            clf_params = self.params.params
        except AttributeError:
            pass

        if clf in clfs.keys():
            self.clf = clfs[clf]
            self.clf_name = clf
        else:
            raise Exception("Model {} not defined in model.py".format(clf))

        # TODO add setting params for clf


        self.clf.fit(X_train, y_train)
        logging.info("finished classifier training")

    def predict(self, X_test, y_test):
        if self.params.metric == "auroc":
            y_proba = self.clf.predict_proba(X_test)[:, 1]
            auroc = roc_auc_score(y_test, y_proba)
            self.result = auroc
            return auroc
        elif self.params.metric == "macro_f1":
            y_pred = self.clf.predict(X_test)
            _, _, f1, _ = (
                precision_recall_fscore_support(y_test, y_pred, average='macro'))
            self.result = f1
            return f1
        else:
            error_message = "Metric {} is not supported in {} yet".format(self.params.metric, self.clf_name)
            logging.error(error_message)
            raise ValueError(error_message)



from imblearn.over_sampling import SMOTE, RandomOverSampler
from imblearn.under_sampling import RandomUnderSampler, OneSidedSelection, CondensedNearestNeighbour
from imblearn.combine import SMOTEENN
from AutoImblearn.components.resamplers import RunSmoteSampler, RunImblearnSampler


import logging
import numpy as np
import pandas as pd

# TODO add back after debug
# from ..components.resamplers.mwmote import MWMOTE

rsps = {
    'rus': RunImblearnSampler(model='rus'),
    'ros': RunImblearnSampler(model='ros'),
    'smote': RunImblearnSampler(model='smote'),
    'mwmote': RunSmoteSampler(model='mwmote'),
}

# rsps = {
#     'rus': RandomUnderSampler(random_state=42),
#     'ros': RandomOverSampler(random_state=42),
#     'smote': SMOTE(random_state=42),
#     # 'mwmote': MWMOTE(random_state=42),
#
#     # 'combined': SMOTEENN(random_state=42, n_jobs=-1),
# }

def value_counter(Y: np.ndarray):
    values, counts = np.unique(Y, return_counts=True)
    for value, count in zip(values, counts):
        dist = count / Y.shape[0] * 100
        logging.info("\t\t Class={}, n={},\t ({:.2f}%)".format(value, count, dist))


class CustomResamplar:
    """ Resamplar oject to re-sample imbalanced data
        Args:
            X (np.ndarray):
            Y (np.ndarray):
    """

    def __init__(self, X: np.ndarray, Y: np.ndarray):

        self.X = X
        self.Y = Y

    def need_resample(self, samratio=None):
        """ Test if need re-sample
        """
        # If nothing is given, use default settings in the models
        if samratio is None:
            return True

        _, counts = np.unique(self.Y, return_counts=True)
        ratio = counts[1] / counts[0]

        return ratio < samratio

    def resample(self, args,  rsp=None, ratio=None):
        logging.info("\t Before Re-Sampling")
        value_counter(self.Y)
        if rsp in rsps.keys():
            resampler = rsps[rsp]
            if ratio is None:
                pass
            elif ratio is not None and hasattr(resampler, 'set_params'):
                resampler.set_params(**{"sampling_strategy": ratio})
            else:
                raise ValueError("can't set resampling ratio for {}".format(rsp))
        else:
            raise ValueError("Re-sampling method {} is not defined".format(rsp))

        self.X, self.Y = resampler.fit_resample(self.X, self.Y)
        logging.info("\t After Re-Sampling")
        value_counter(self.Y)

        return self.X, self.Y



#import imbalanced_databases as imbd
from collections import Counter
import numpy as np

def get_data(name, val_ratio=0.2, test_raito=0.2, undersample_ratio=100):
    from sklearn.datasets import fetch_openml
    from imblearn.datasets import make_imbalance
    from sklearn.impute import SimpleImputer
    from sklearn.preprocessing import StandardScaler
    from sklearn.preprocessing import OneHotEncoder
    from sklearn.pipeline import make_pipeline
    from sklearn.compose import make_column_transformer
    from sklearn.compose import make_column_selector as selector
    from sklearn import preprocessing
    version=1
    imp_file_names = ["imp_ii_.p", "imp_median_.p", "imp_knn_.p", "imp_gain_.p", "imp_MIRACLE_.p", "imp_MIWAE_.p"]
    if name in imp_file_names:
        import pickle
        import os
        import pandas as pd
        file =  open(os.path.join("../../../..", "data", "interim", name), "rb")
        X, y = pickle.load(file)

        X = X.to_numpy()
        file.close()
    else:

        X, y = fetch_openml(name, version=version, as_frame=True, return_X_y=True)
        num_pipe = make_pipeline(
            StandardScaler(), SimpleImputer(strategy="mean", add_indicator=True)
        )
        cat_pipe = make_pipeline(
            SimpleImputer(strategy="constant", fill_value="missing"),
            OneHotEncoder(handle_unknown="ignore"),
        )
        preprocessor_linear = make_column_transformer(
            (num_pipe, selector(dtype_include="number")),
            (cat_pipe, selector(dtype_include="category")),
            n_jobs=2,
        )
        X = preprocessor_linear.fit_transform(X)

        le = preprocessing.LabelEncoder()
        le.fit(y)
        y = le.transform(y)

        if type(X) != np.ndarray:
            X = X.todense()

    print("========== Dataset Info ==========")
    print("X shape:", X.shape, "y shape:", y.shape)
    print("Counts:", Counter(y))
    print("==================================")
    size = X.shape[0]
    # Shuffle
    indices = [i for i in range(size)]
    np.random.shuffle(indices)
    val_idx, test_idx, train_idx = indices[:int(size*val_ratio)], indices[int(size*val_ratio):int(size*(val_ratio+test_raito))], indices[int(size*(val_ratio+test_raito)):]

    train_X, val_X, test_X = X[train_idx], X[val_idx], X[test_idx]
    train_y, val_y, test_y = y[train_idx], y[val_idx], y[test_idx]

    if name != "2022-10-06 NHANES 1999-2010.csv":
        class_counts = Counter(train_y)
        train_X, train_y = make_imbalance(
            train_X,
            train_y,
            sampling_strategy={
                min(class_counts, key=class_counts.get): max(class_counts.values()) // undersample_ratio,
            },
        )
        class_counts = Counter(val_y)
        val_X, val_y = make_imbalance(
            val_X,
            val_y,
            sampling_strategy={
                min(class_counts, key=class_counts.get): max(class_counts.values()) // undersample_ratio,
            },
        )
        class_counts = Counter(test_y)
        test_X, test_y = make_imbalance(
            test_X,
            test_y,
            sampling_strategy={
                min(class_counts, key=class_counts.get): max(class_counts.values()) // undersample_ratio,
            },
        )

    print("========== Training Info =========")
    print("train X shape:", train_X.shape, "train y shape:", train_y.shape)
    print("Counts:", Counter(train_y))
    print("==================================")

    return train_X, train_y, val_X, val_y, test_X, test_y




import copy
import csv
import datetime
import json
import logging
import os
import time
import collections
from typing import Dict, List

import torch
import torch.nn.functional as F
from torch import multiprocessing as mp

Buffers = Dict[str, List[torch.Tensor]]

def _format_frame(frame):
    src_features = torch.from_numpy(frame["src_features"]).float()
    src_features = src_features.view((1, 1) + src_features.shape)  # (...) -> (T,B,...).
    neighbor_features = torch.from_numpy(frame["neighbor_features"]).float()
    neighbor_features = neighbor_features.view((1, 1) + neighbor_features.shape)  # (...) -> (T,B,...).
    return src_features, neighbor_features


class Environment:
    def __init__(self, gym_env):
        self.gym_env = gym_env
        self.episode_return = None
        self.episode_step = None

    def initial(self):
        initial_reward = torch.zeros(1, 1)
        #initial_last_action = torch.zeros(1, 1, dtype=torch.int64)
        self.episode_return = torch.zeros(1, 1)
        self.episode_step = torch.zeros(1, 1, dtype=torch.int32)
        initial_done = torch.ones(1, 1, dtype=torch.uint8)
        initial_src_features, initial_neighbor_features  = _format_frame(self.gym_env.reset())

        return dict(
            src_features=initial_src_features,
            neighbor_features=initial_neighbor_features,
            reward=initial_reward,
            done=initial_done,
            episode_return=self.episode_return,
            episode_step=self.episode_step,
            #last_action=initial_last_action,
        )

    def step(self, action):
        frame, reward, done, info = self.gym_env.step(action)
        self.episode_step += 1
        self.episode_return += reward
        episode_step = self.episode_step
        episode_return = self.episode_return
        if done:
            frame = self.gym_env.reset()
            self.episode_return = torch.zeros(1, 1)
            self.episode_step = torch.zeros(1, 1, dtype=torch.int32)

        src_features, neighbor_features  = _format_frame(frame)
        reward = torch.tensor(reward).view(1, 1)
        done = torch.tensor(done).view(1, 1)

        return dict(
            src_features=src_features,
            neighbor_features=neighbor_features,
            reward=reward,
            done=done,
            episode_return=episode_return,
            episode_step=episode_step,
            #last_action=action,
        ), info

    def close(self):
        self.gym_env.close()

def get_batch(
    flags,
    free_queue: mp.SimpleQueue,
    full_queue: mp.SimpleQueue,
    buffers: Buffers,
    indices=None
):
    if indices is None:
        indices = [full_queue.get() for _ in range(flags.batch_size)]
    batch = {
        key: torch.stack([buffers[key][m] for m in indices], dim=1) for key in buffers
    }
    for m in indices:
        free_queue.put(m)

    batch = {k: t.to(device=flags.device, non_blocking=True) for k, t in batch.items()}

    return batch

def compute_baseline_loss(advantages):
    return 0.5 * torch.sum(advantages ** 2)

def compute_entropy_loss(logits):
    """Return the entropy loss, i.e., the negative entropy of the policy."""
    policy = F.softmax(logits, dim=-1)
    log_policy = F.log_softmax(logits, dim=-1)
    return torch.sum(policy * log_policy)

def compute_policy_gradient_loss(logits, actions, advantages):
    cross_entropy = F.nll_loss(
        F.log_softmax(torch.flatten(logits, 0, 1), dim=-1),
        target=torch.flatten(actions, 0, 1),
        reduction="none",
    )
    cross_entropy = cross_entropy.view_as(advantages)
    return torch.sum(cross_entropy * advantages.detach())

def gather_metadata() -> Dict:
    date_start = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")
    # Gathering git metadata.
    try:
        import git

        try:
            repo = git.Repo(search_parent_directories=True)
            git_sha = repo.commit().hexsha
            git_data = dict(
                commit=git_sha,
                branch=None if repo.head.is_detached else repo.active_branch.name,
                is_dirty=repo.is_dirty(),
                path=repo.git_dir,
            )
        except git.InvalidGitRepositoryError:
            git_data = None
    except ImportError:
        git_data = None
    # Gathering slurm metadata.
    if "SLURM_JOB_ID" in os.environ:
        slurm_env_keys = [k for k in os.environ if k.startswith("SLURM")]
        slurm_data = {}
        for k in slurm_env_keys:
            d_key = k.replace("SLURM_", "").replace("SLURMD_", "").lower()
            slurm_data[d_key] = os.environ[k]
    else:
        slurm_data = None
    return dict(
        date_start=date_start,
        date_end=None,
        successful=False,
        git=git_data,
        slurm=slurm_data,
        env=os.environ.copy(),
    )

class FileWriter:
    def __init__(
        self,
        xpid: str = None,
        xp_args: dict = None,
        rootdir: str = "~/logs",
        symlink_to_latest: bool = True,
    ):
        if not xpid:
            # Make unique id.
            xpid = "{proc}_{unixtime}".format(
                proc=os.getpid(), unixtime=int(time.time())
            )
        self.xpid = xpid
        self._tick = 0

        # Metadata gathering.
        if xp_args is None:
            xp_args = {}
        self.metadata = gather_metadata()
        # We need to copy the args, otherwise when we close the file writer
        # (and rewrite the args) we might have non-serializable objects (or
        # other unwanted side-effects).
        self.metadata["args"] = copy.deepcopy(xp_args)
        self.metadata["xpid"] = self.xpid

        formatter = logging.Formatter("%(message)s")
        self._logger = logging.getLogger("logs/out")

        # To stdout handler.
        shandle = logging.StreamHandler()
        shandle.setFormatter(formatter)
        self._logger.addHandler(shandle)
        self._logger.setLevel(logging.INFO)

        rootdir = os.path.expandvars(os.path.expanduser(rootdir))
        # To file handler.
        self.basepath = os.path.join(rootdir, self.xpid)
        if not os.path.exists(self.basepath):
            self._logger.info("Creating log directory: %s", self.basepath)
            os.makedirs(self.basepath, exist_ok=True)
        else:
            self._logger.info("Found log directory: %s", self.basepath)

        if symlink_to_latest:
            # Add 'latest' as symlink unless it exists and is no symlink.
            symlink = os.path.join(rootdir, "latest")
            try:
                if os.path.islink(symlink):
                    os.remove(symlink)
                if not os.path.exists(symlink):
                    os.symlink(self.basepath, symlink)
                    self._logger.info("Symlinked log directory: %s", symlink)
            except OSError:
                # os.remove() or os.symlink() raced. Don't do anything.
                pass

        self.paths = dict(
            msg="{base}/out.log".format(base=self.basepath),
            logs="{base}/logs.csv".format(base=self.basepath),
            fields="{base}/fields.csv".format(base=self.basepath),
            meta="{base}/meta.json".format(base=self.basepath),
        )

        self._logger.info("Saving arguments to %s", self.paths["meta"])
        if os.path.exists(self.paths["meta"]):
            self._logger.warning(
                "Path to meta file already exists. " "Not overriding meta."
            )
        else:
            self._save_metadata()

        self._logger.info("Saving messages to %s", self.paths["msg"])
        if os.path.exists(self.paths["msg"]):
            self._logger.warning(
                "Path to message file already exists. " "New data will be appended."
            )

        fhandle = logging.FileHandler(self.paths["msg"])
        fhandle.setFormatter(formatter)
        self._logger.addHandler(fhandle)

        self._logger.info("Saving logs data to %s", self.paths["logs"])
        self._logger.info("Saving logs' fields to %s", self.paths["fields"])
        self.fieldnames = ["_tick", "_time"]
        if os.path.exists(self.paths["logs"]):
            self._logger.warning(
                "Path to log file already exists. " "New data will be appended."
            )
            # Override default fieldnames.
            with open(self.paths["fields"], "r") as csvfile:
                reader = csv.reader(csvfile)
                lines = list(reader)
                if len(lines) > 0:
                    self.fieldnames = lines[-1]
            # Override default tick: use the last tick from the logs file plus 1.
            with open(self.paths["logs"], "r") as csvfile:
                reader = csv.reader(csvfile)
                lines = list(reader)
                # Need at least two lines in order to read the last tick:
                # the first is the csv header and the second is the first line
                # of data.
                if len(lines) > 1:
                    self._tick = int(lines[-1][0]) + 1

        self._fieldfile = open(self.paths["fields"], "a")
        self._fieldwriter = csv.writer(self._fieldfile)
        self._logfile = open(self.paths["logs"], "a")
        self._logwriter = csv.DictWriter(self._logfile, fieldnames=self.fieldnames)

    def log(self, to_log: Dict, tick: int = None, verbose: bool = False) -> None:
        if tick is not None:
            raise NotImplementedError
        else:
            to_log["_tick"] = self._tick
            self._tick += 1
        to_log["_time"] = time.time()

        old_len = len(self.fieldnames)
        for k in to_log:
            if k not in self.fieldnames:
                self.fieldnames.append(k)
        if old_len != len(self.fieldnames):
            self._fieldwriter.writerow(self.fieldnames)
            self._logger.info("Updated log fields: %s", self.fieldnames)

        if to_log["_tick"] == 0:
            self._logfile.write("# %s\n" % ",".join(self.fieldnames))

        if verbose:
            self._logger.info(
                "LOG | %s",
                ", ".join(["{}: {}".format(k, to_log[k]) for k in sorted(to_log)]),
            )

        self._logwriter.writerow(to_log)
        self._logfile.flush()

    def close(self, successful: bool = True) -> None:
        self.metadata["date_end"] = datetime.datetime.now().strftime(
            "%Y-%m-%d %H:%M:%S.%f"
        )
        self.metadata["successful"] = successful
        self._save_metadata()

        for f in [self._logfile, self._fieldfile]:
            f.close()

    def _save_metadata(self) -> None:
        with open(self.paths["meta"], "w") as jsonfile:
            json.dump(self.metadata, jsonfile, indent=4, sort_keys=True)

VTraceFromLogitsReturns = collections.namedtuple(
    "VTraceFromLogitsReturns",
    [
        "vs",
        "pg_advantages",
        "log_rhos",
        "behavior_action_log_probs",
        "target_action_log_probs",
    ],
)

VTraceReturns = collections.namedtuple("VTraceReturns", "vs pg_advantages")

def action_log_probs(policy_logits, actions):
    return -F.nll_loss(
        F.log_softmax(torch.flatten(policy_logits, 0, -2), dim=-1),
        torch.flatten(actions),
        reduction="none",
    ).view_as(actions)


def from_logits(
    behavior_policy_logits,
    target_policy_logits,
    actions,
    discounts,
    rewards,
    values,
    bootstrap_value,
    clip_rho_threshold=1.0,
    clip_pg_rho_threshold=1.0,
):
    """V-trace for softmax policies."""

    target_action_log_probs = action_log_probs(target_policy_logits, actions)
    behavior_action_log_probs = action_log_probs(behavior_policy_logits, actions)
    log_rhos = target_action_log_probs - behavior_action_log_probs
    vtrace_returns = from_importance_weights(
        log_rhos=log_rhos,
        discounts=discounts,
        rewards=rewards,
        values=values,
        bootstrap_value=bootstrap_value,
        clip_rho_threshold=clip_rho_threshold,
        clip_pg_rho_threshold=clip_pg_rho_threshold,
    )
    return VTraceFromLogitsReturns(
        log_rhos=log_rhos,
        behavior_action_log_probs=behavior_action_log_probs,
        target_action_log_probs=target_action_log_probs,
        **vtrace_returns._asdict(),
    )


@torch.no_grad()
def from_importance_weights(
    log_rhos,
    discounts,
    rewards,
    values,
    bootstrap_value,
    clip_rho_threshold=1.0,
    clip_pg_rho_threshold=1.0,
):
    """V-trace from log importance weights."""
    with torch.no_grad():
        rhos = torch.exp(log_rhos)
        if clip_rho_threshold is not None:
            clipped_rhos = torch.clamp(rhos, max=clip_rho_threshold)
        else:
            clipped_rhos = rhos

        cs = torch.clamp(rhos, max=1.0)
        # Append bootstrapped value to get [v1, ..., v_t+1]
        values_t_plus_1 = torch.cat(
            [values[1:], torch.unsqueeze(bootstrap_value, 0)], dim=0
        )
        deltas = clipped_rhos * (rewards + discounts * values_t_plus_1 - values)

        acc = torch.zeros_like(bootstrap_value)
        result = []
        for t in range(discounts.shape[0] - 1, -1, -1):
            acc = deltas[t] + discounts[t] * cs[t] * acc
            result.append(acc)
        result.reverse()
        vs_minus_v_xs = torch.stack(result)

        # Add V(x_s) to get v_s.
        vs = torch.add(vs_minus_v_xs, values)

        # Advantage for policy gradient.
        broadcasted_bootstrap_values = torch.ones_like(vs[0]) * bootstrap_value
        vs_t_plus_1 = torch.cat(
            [vs[1:], broadcasted_bootstrap_values.unsqueeze(0)], dim=0
        )
        if clip_pg_rho_threshold is not None:
            clipped_pg_rhos = torch.clamp(rhos, max=clip_pg_rho_threshold)
        else:
            clipped_pg_rhos = rhos
        pg_advantages = clipped_pg_rhos * (rewards + discounts * vs_t_plus_1 - values)

        # Make sure no gradients backpropagated through the returned values.
        return VTraceReturns(vs=vs, pg_advantages=pg_advantages)




import os
import shutil
import re

from flask import Flask, jsonify, request
import pandas as pd
from sklearn.model_selection import train_test_split
from autosklearn.metrics import f1_macro, roc_auc

import autosklearn.classification

app = Flask(__name__)

params = {}

# POST function to set training parameters
@app.route('/set', methods=['POST'])
def set_params():
    """ Set training parameters """
    data = request.get_json()
    global params
    for key, value in data.items():
        params[key] = value
    if 'metric' not in params or 'dataset' not in params or 'target' not in params:
        raise Exception("data not complete, need to include metric, dataset, and target")

    return jsonify(params), 201

# GET function to retrieve all books
@app.route('/results/<dataset_name>', methods=['GET'])
def get_result(dataset_name: str):
    """ Train the auto-sklearn and return the result in json format """
    global params
    shutil.rmtree('/tmp', ignore_errors=True)
    null_values = ['', ' ']
    data_path = os.path.join("/data/raw", params['dataset'])
    if not os.path.isfile(data_path):
        return jsonify({"result": "file not exist"}, 200)

    file_type = re.search("[^\.]*$", data_path).group()
    if file_type == "csv":
        df = pd.read_csv(data_path, na_values=null_values)
    elif file_type in ["xls,", "xlsx,", "xlsm,", "xlsb,", "odf,", "ods,", "odt"]:
        df = pd.read_excel(data_path, na_values=null_values)
    else:
        return jsonify({"error": "file type of {} not supported".format(dataset_name)}), 400

    headers = list(df.columns.values)
    target = params["target"]
    headers.remove(target)

    X = df[headers]
    y = df[target]

    # The AutoML part
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, stratify=y, test_size=0.1
    )
    automl = autosklearn.classification.AutoSklearnClassifier(
        time_left_for_this_task=120,
        per_run_time_limit=30,
        tmp_folder="/tmp/autosklearn_classification_example_tmp",
        metric=roc_auc,
        resampling_strategy="cv",
        resampling_strategy_arguments={"folds": 10}
    )

    if params["metric"] == "auroc":
        automl.set_params(**{"metric": roc_auc})
    elif params["metric"] == "macro_f1":
        automl.set_params(**{"metric": f1_macro})
    else:
        raise Exception("Evaluation metric {} is not yet supported".format(params["metric"]))

    # Train
    automl.fit(X_train, y_train)

    # Predict
    if params["metric"] == "auroc":
        y_proba = automl.predict_proba(X_test)
        result = roc_auc(y_test, y_proba)
    elif params["metric"] == "macro_f1":
        y_pred = automl.predict(X_test)
        result = f1_macro(y_test, y_pred)
    else:
        raise Exception("Evaluation metric {} is not yet supported".format(params["metric"]))

    return jsonify({"result": result}), 200

if __name__ == '__main__':
    app.run(host='0.0.0.0', port=8080, debug=True)


import numpy as np
import gym
from sklearn.metrics import f1_score, matthews_corrcoef, roc_auc_score
from sklearn.metrics.pairwise import euclidean_distances

class Env:
    def __init__(
        self,
        train_X,
        train_y,
        val_X,
        val_y,
        clf,
        metric,
    ):
        self.train_X = train_X
        self.train_y = train_y
        self.val_X = val_X
        self.val_y = val_y
        self.clf = clf
        self.metric = metric

        unique, counts = np.unique(self.train_y, return_counts=True)
        self.min_label = unique[0] if counts[0] < counts[1] else unique[1]
        self.train_X_min = self.train_X[self.train_y==self.min_label]
        
    def step(self, samples):
        if len(samples) > 0:
            train_X = np.vstack([self.train_X, samples])
            train_y = np.hstack([self.train_y, np.hstack([self.min_label]*len(samples))])
        else:
            train_X, train_y = self.train_X, self.train_y
        self.clf.fit(train_X, train_y)
        pred_y = self.clf.predict(self.val_X)
        if self.metric == "macro_f1":
            result = f1_score(self.val_y, pred_y, average="macro")
        elif self.metric == "mcc":
            result = matthews_corrcoef(self.val_y, pred_y)
        elif self.metric == "auroc":
            if hasattr(self.clf, 'predict_proba'):
                pred_y_proba = self.clf.predict_proba(self.val_X)[:, 1]
            else:
                pred_y_proba = pred_y
            result = roc_auc_score(self.val_y, pred_y_proba)
        return result

class GymWrapper(gym.Env):
    def __init__(self, env, flags, neighbors=None):
        self._env = env
        self.X_min = self._env.train_X_min
        self.length = len(self.X_min)
        self.num_instance_specific_actions = flags.num_instance_specific_actions
        self.feature_dim = self.X_min.shape[1]

        self.count = np.zeros(self.length, dtype=np.int32)
        count_features = np.zeros((self.length, 10))
        self.instance_features = np.concatenate((count_features, self.X_min), axis=1)
        if neighbors is None:
            self.neighbors = flags.neighbors
        else:
            self.neighbors = neighbors

        # Precompute the pairwise diffs.
        self.diffs = self.X_min[np.newaxis,:,:] - self.X_min[:,np.newaxis,:]
        self.diffs = self.diffs.reshape(-1,self.diffs.shape[-1])

        # Find neighbors
        self.distances = euclidean_distances(self.X_min, self.X_min)
        self.nearest_neighbors = np.argsort(self.distances)

        # For gym
        self.observation_space = gym.spaces.Dict({
            "src_features": gym.spaces.Box(low=-np.inf, high=np.inf, shape=(self.feature_dim+10,), dtype=np.float32),
            "neighbor_features": gym.spaces.Box(low=-np.inf, high=np.inf, shape=(self.neighbors, self.feature_dim+10), dtype=np.float32),
        })
        self.action_space = gym.spaces.Dict({
            "dst_index": gym.spaces.Box(low=-np.inf, high=np.inf, shape=(1,), dtype=np.int32),
            "ratio": gym.spaces.Box(low=-np.inf, high=np.inf, shape=(1,), dtype=np.float32),
            "next": gym.spaces.Box(low=-np.inf, high=np.inf, shape=(1,), dtype=np.int32),
        })

    def seed(self, seed):
        self._env.seed = seed

    def reset(self):
        self.src_indices = []
        self.dst_indices = []
        self.ratio = []
        self.cur_index = 0
        self.count[:] = 0
        self.instance_features[:,0] = 1.0
        self.instance_features[:,1:10] = 0.0

        return self._get_obs()

    def step(self, action):
        # actions contain src, dst, ratio
        if action["dst_index"] >= 0:
            self.src_indices.append(self.cur_index)
            self.dst_indices.append(self.nearest_neighbors[self.cur_index][action["dst_index"]])
            dst_index = self.nearest_neighbors[self.cur_index][action["dst_index"]]
            self.count[self.cur_index] += 1
            if self.count[self.cur_index] % 20 == 0:
                col = min(self.count[self.cur_index] // 20, 9)
                if col > 0:
                    self.instance_features[self.cur_index][col-1] = 0.0
                self.instance_features[self.cur_index][col] = 1.0

            self.count[dst_index] += 1
            if self.count[dst_index] % 20 == 0:
                col = min(self.count[dst_index] // 20, 9)
                if col > 0:
                    self.instance_features[dst_index][col-1] = 0.0
                self.instance_features[dst_index][col] = 1.0

            self.ratio.append(action["ratio"])

        if action["next"]:
            self.cur_index += 1

        obs = self._get_obs()

        if self.cur_index < self.length:
            return obs, 0, False, {}
        else:
            samples = {
                "src_indices": np.array(self.src_indices),
                "dst_indices": np.array(self.dst_indices),
                "ratio": np.array(self.ratio),
            }
            reward = self.get_reward(samples)
            return obs, reward, True, {"samples": samples}

    def get_reward(self, samples):
        # Generating samples
        if len(samples["src_indices"]) == 0:
            samples = []
        else:
            src_indices = samples["src_indices"]
            dst_indices = samples["dst_indices"]
            ratio = samples["ratio"]
            flat_indices = src_indices * self.length + dst_indices
            samples = self.X_min[src_indices] + ratio.reshape(ratio.shape[0],1) * self.diffs[flat_indices]

        return self._env.step(samples)

    def _get_obs(self):
        if self.cur_index >= self.length:
            return None
        else:
            return {
                "src_features": self.instance_features[self.cur_index],
                "neighbor_features": self.instance_features[self.nearest_neighbors[self.cur_index][:self.neighbors]],
            }



import json
import os
import numpy as np
import logging
import docker
import requests
from time import sleep


import sys
sys.path.append("..")
from ...processing.utils import DATA_VOLUME_PATH


class Arguments:
    def __init__(self):
        self.dataset = "nhanes.csv"
        self.metric = "auroc"
        self.target = "Status"

        self.device = "cpu"
        self.cuda = "0"

        self.val_ratio = 0.1,
        self.test_raito = 0.1,


class RunAutoRSP:
    def __init__(self):
        self.flags = Arguments()
        os.environ["CUDA_VISIBLE_DEVICES"] = self.flags.cuda
        self.supported_metrics = ["auroc, macro_f1"]
        self.result = None

    def fit(self, X_train: np.ndarray, y_train: np.ndarray, X_test: np.ndarray, y_test: np.ndarray, metric=None):
        self.flags.metric = metric

        import pandas as pd
        train_data = np.hstack((X_train, y_train.reshape(-1, 1)))
        test_data = np.hstack((X_test, y_test.reshape(-1, 1)))
        train_df = pd.DataFrame(train_data,columns=[f'feature_{i}' for i in range(X_train.shape[1])] + ['target'])
        test_df = pd.DataFrame(test_data,columns=[f'feature_{i}' for i in range(X_test.shape[1])] + ['target'])
        train_df.to_csv('../../data/interim/autorsp_train.csv', index=False)
        test_df.to_csv('../../data/interim/autorsp_test.csv', index=False)

        client = docker.from_env()

        # Get the image build
        image_name = 'autorsp:version1.0'
        try:
            client.images.get(image_name)
            logging.info('found image')
        except docker.errors.ImageNotFound:
            logging.info("Building AutoResampling 1.0 image")
            client.images.build(path="autorsp/", tag=image_name, nocache=True)
        # Create the container
        volume1 = os.path.abspath("autorsp")
        volume2 = os.path.abspath(DATA_VOLUME_PATH)
        container_name = "autorsp-flask"

        logging.info('Creating AutoResampling container')

        try:
            container = client.containers.get(container_name)
            logging.info('found container')
        except:
            container = client.containers.run(
                name=container_name,
                image=image_name,
                ports={"8083": 8083},
                volumes=['{}:/code'.format(volume1),
                         '{}:/data'.format(volume2)],
                entrypoint="Rscript /code/app/autorsp.R",
                detach=True,
            )
        timeout = 120
        stop_time = 3
        elapsed_time = 0
        logging.info('waiting container to be ready')
        print(container.status)
        while container.status != 'running' and elapsed_time < timeout:
            if container.status == 'exited':
                raise Exception("AutoResampling docker container build error")
            sleep(stop_time)
            elapsed_time += stop_time
            container = client.containers.get(container_name)
            continue
        print(container.status)

        sleep(5)
        # logging.info('Getting result from REST API')
        get_url = "http://0.0.0.0:8083/result?metric={}&target=target".format(metric)
        print(get_url)
        response_API = requests.get(get_url)
        self.result = response_API.json()
        print(self.result)

        container.remove(force=True, v=True)

    def predict(self, X_test: np.ndarray = None, y_test: np.ndarray = None):
        return self.result

if __name__ == "__main__":
    import pickle
    from src.utils import  Samplar
    file_path = '../../data/interim/imp_gain_.p'
    with open(file_path, 'rb') as file:
        data = pickle.load(file)
    train_sampler = Samplar(np.array(data[0]), np.array(data[1]))
    for X_train, y_train, X_test, y_test in train_sampler.apply_kfold(10):
        run = RunAutoRSP()
        run.fit(X_train, y_train, X_test, y_test, metric="macro_f1")
        break


import logging

import numpy as np

from xgboost import XGBClassifier

from AutoImblearn.components.api import BaseEstimatorAPI

clfs = {
    "ensemble": XGBClassifier(learning_rate=1.0, max_depth=10, min_child_weight=15, n_estimators=100, n_jobs=1, subsample=0.8, verbosity=0),
}

class RunSklearnClassifierAPI(BaseEstimatorAPI):
    def get_hyperparameter_search_space(self):
        return {
            "learning_rate": {
                "type": "float",
                "min": 0.01,
                "max": 1.0,
                "default": 1.0,
                "log_scale": True
            },
            "max_depth": {
                "type": "int",
                "min": 1,
                "max": 20,
                "default": 10
            },
            "min_child_weight": {
                "type": "int",
                "min": 1,
                "max": 30,
                "default": 15
            },
            "n_estimators": {
                "type": "int",
                "min": 10,
                "max": 1000,
                "default": 100
            },
            "subsample": {
                "type": "float",
                "min": 0.5,
                "max": 1.0,
                "default": 0.8
            },
            "n_jobs": {
                "type": "int",
                "min": 1,
                "max": 16,
                "default": 1
            },
            "verbosity": {
                "type": "categorical",
                "choices": [0, 1, 2, 3],
                "default": 0
            }
        }

    def fit(self, args, X_train, y_train, X_test, y_test):
        clf = self.params.model
        params = self.dict_to_namespace()
        params.metric = args.metric
        params.model = args.model

        if clf in clfs.keys():
            self.clf = clfs[clf]
            self.clf_name = clf
        else:
            raise Exception("Model {} not defined in model.py".format(clf))

        self.clf.fit(X_train, Y_train)

        # size = X_train.shape[0]
        # indices = np.arange(size)
        # np.random.shuffle(indices)
        #
        # val_idx = indices[:int(size * args.val_ratio)]
        # train_idx = indices[int(size * args.val_ratio):]
        #
        # train_X, val_X = X_train[train_idx], X_train[val_idx]
        # train_y, val_y = y_train[train_idx], y_train[val_idx]

        logging.info("finished parameter setting")

        params.ratio_map = [0.0, 0.25, 0.5, 0.75, 1.0]
        clf = get_clf(params.clf)
        return train(params, train_X, train_y, val_X, val_y, X_test, y_test, clf)

    def predict(self, dataset_name: str):
        pass



"""
AutoImblearn

The automated machine learning system for imbalanced data
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='AutoImblearn',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version="0.1.12",

    description='Automated machine learning system for imbalanced medical data',
    long_description=long_description,
    long_description_content_type="text/markdown",

    # The project's main homepage.
    # url='https://github.com/Wanghongkua/Auto-Imblearn2',

    # Author details
    author='Hank Wang',
    author_email='hankwang1991@gmail.com',

    # Choose your license
    license='BSD 3-Clause License',

    # Python Versions
    python_requires=">=3.9",

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        "Programming Language :: Python :: 3",
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],

    # What does your project relate to?
    keywords=['medical analysis','automated machine learning'],

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        "docker",
        "joblib",
        "pandas",
        "scikit-learn",
        "pydantic",
    ],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]

    extras_require={
        "web": [
            "matplotlib",
            "seaborn",
        ],
        "imputer": [
            "hyperimpute",  # installed only when requested
        ],
        "resampler": [
            "smote_variants",
            "imbalanced-learn",
        ],
        "r_model": [
            "rpy2",
        ],
        'dev': [
            'check-manifest',
        ],
    },

)



import os
import pickle
from collections import Counter
from typing import Dict

from LDAword2vec import LDAword2vec
# from runDoc2vec import Doc2Vec
from ...processing.utils import DataLoader


def read_txt_files_to_corpus(model_folder_path):
    corpus = {}

    # Iterate over all files in the specified folder
    for filename in os.listdir(model_folder_path):
        # Check if the file has a .txt extension
        if filename.endswith('.txt'):
            # Get the file path
            file_path = os.path.join(model_folder_path, filename)

            # Open the file and read the content
            with open(file_path, 'r', encoding='utf-8') as file:
                content = file.read()

            # Remove the file extension (.txt) from the filename to use as the key
            file_key = os.path.splitext(filename)[0]

            # Add the content to the corpus with the filename as the key
            corpus[file_key] = content

    return corpus


def count_top_models(result_file):
    """ Find the top models from exhaustive search results """
    with open(result_file, 'rb') as f:
        data = pickle.load(f)

    # formatted_data = json.dumps(data, indent=4)
    # pprint(formatted_data)
    #
    #
    # try:
    #     data['MIRACLE'] = data['miracle']
    #     del data['miracle']
    #     data['MIWAE'] = data['miwae']
    #     del data['miwae']
    #     with open(result_file, "wb") as f:
    #         pickle.dump(data, f)
    # except KeyError:
    #     pass


    # Validate the value key pair
    # for imputer, resamplers in data.items():
    #     if not data[imputer]:
    #         del data[imputer]
    #         print("{} is deleted".format(imputer))
    #     for resampler, classifiers in resamplers.items():
    #         if not resamplers[resampler]:
    #             del resamplers[resampler]
    #             print("{} {} is deleted".format(imputer, resampler))
    #         for classifier, value in classifiers.items():
    #             if not classifiers[classifier]:
    #                 del classifiers[classifier]
    #                 print("{} {} {} is deleted".format(imputer, resampler, classifier))

    # Initialize counters
    classifier_counter = Counter()
    resampler_counter = Counter()
    imputer_counter = Counter()

    classifier_names = list(next(iter(next(iter(data.values())).values())).keys())

    # Fix imputer and resampler -> Find best classifier
    for imputer, resamplers in data.items():
        for resampler, classifiers in resamplers.items():
            best_classifier = max(classifiers, key=classifiers.get)
            classifier_counter[best_classifier] += 1

    # Fix imputer and classifier -> Find best resampler
    for imputer, resamplers in data.items():
        for classifier in classifier_names:
            best_resampler = max(resamplers, key=lambda r: resamplers[r][classifier])
            resampler_counter[best_resampler] += 1

    # Fix resampler and classifier -> Find best imputer
    for resampler in list(data.values())[0]:
        for classifier in classifier_names:
            best_imputer = max(data, key=lambda i: data[i][resampler][classifier])
            imputer_counter[best_imputer] += 1

    return {'imputer':imputer_counter, 'resampler': resampler_counter, 'classifier': classifier_counter}


def get_query(dataset_name):
    """get the dataset description based on dataset name"""
    dataset_dp_path = os.path.join("../../../..", "data", 'datasets', dataset_name + ".txt")
    if os.path.isfile(dataset_dp_path):
        with open(dataset_dp_path, "r") as f:
            query = f.read()
        return query
    else:
        raise ValueError(f"Description for dataset {dataset_name} doesn't exist in file path {dataset_dp_path}")


def find_best_pipes(dataset_names, metrics):
    # Find the best pipes for all datasets
    # result = {key: {dataset: {} for dataset in dataset_names} for key in metrics}
    result = {dataset: {key: {} for key in metrics} for dataset in dataset_names}

    for dataset in dataset_names:
        data_folder_path = os.path.join("../../../..", 'data', 'processed', dataset)
        if not os.path.isdir(data_folder_path):
            continue

        for file in os.listdir(data_folder_path):
            if file.endswith(".p"):
                for metric in metrics:
                    if metric in file:
                        result_file = os.path.join(data_folder_path, file)
                        result[dataset][metric] = count_top_models(result_file)
    return result


def get_accuracy(truth: Dict[str, Dict[str, int]], result: Dict):
    """ compare LLM result and ground truth

    : truth : [model type, top count]
    : result: [model type, model ranking]
     """
    accuracy = {}
    for model_type, counts in truth.items():
        total_count = 0
        hit_count = 0
        for model, count in counts.items():
            total_count += count
            if model in result[model_type]:
                hit_count += count
        accuracy[model_type] = hit_count/total_count

    return accuracy


if __name__ == "__main__":
    data_loader = DataLoader(dataset = "pima-indians-diabetes-missing.csv")
    dp = data_loader.get_data_description("Pima")
    # model_filtering = ModelFiltering(dp)
    # print(model_filtering.get_topn("Imputers"))



import json
import os
import numpy as np
import logging
import docker
import requests
from time import sleep

from ...processing.utils import DATA_VOLUME_PATH


class Arguments:
    def __init__(self):
        self.dataset = "nhanes.csv"
        self.metric = "auroc"
        self.target = "Status"

        self.device = "cpu"
        self.cuda = "0"

        self.val_ratio = 0.1,
        self.test_raito = 0.1,


class RunAutoSklearn:
    def __init__(self):
        self.flags = Arguments()
        os.environ["CUDA_VISIBLE_DEVICES"] = self.flags.cuda
        self.supported_metrics = ["auroc, macro_f1"]
        self.result = None

    def fit(self, X_train: np.ndarray, y_train: np.ndarray, metric=None):
        self.flags.metric = metric


        # Start training
        client = docker.from_env()

        # Get the image build
        image_name = 'autosklearn:version1.0'
        try:
            client.images.get(image_name)
            logging.info('found image')
        except docker.errors.ImageNotFound:
            logging.info("Building AutoSklearn 1.0 image")
            client.images.build(path="autosklearn/", tag=image_name, nocache=True)
        # Create the container
        volume1 = os.path.abspath("autosklearn")
        volume2 = os.path.abspath(DATA_VOLUME_PATH)
        container_name = "autosklearn-flask"

        logging.info('Creating AutoSklearn container')

        try:
            container = client.containers.get(container_name)
            # container.remove(force=True, v=True)
            # print("container remove")
            logging.info('found container')
        except:
            container = client.containers.run(
                name=container_name,
                image=image_name,
                ports={"8080": 8080},
                volumes=['{}:/code'.format(volume1),
                         # start container from a container code
                         # '/var/run/docker.sock:/var/run/docker.sock',
                         '{}:/data'.format(volume2)],
                entrypoint="python3 /code/app/app.py",
                detach=True,
            )
        timeout = 120
        stop_time = 3
        elapsed_time = 0
        logging.info('waiting container to be ready')

        while container.status != 'running' and elapsed_time < timeout:
            if container.status == 'exited':
                raise Exception("AutoSklearn docker container build error")
            sleep(stop_time)
            elapsed_time += stop_time
            container = client.containers.get(container_name)
            continue
        # print(container.status)
        # POST -- set parameters
        post_url = 'http://127.0.0.1:8080/set'
        headers =  {"Content-Type":"application/json"}
        response = requests.post(post_url, json.dumps(self.flags.__dict__), headers=headers)
        if response.status_code != 201:
            raise Exception("There is an error in setting AutoSklearn parameters")

        # GET -- get result
        logging.info('Getting result from REST API')
        get_url = 'http://127.0.0.1:8080/results/' + self.flags.dataset
        response_API = requests.get(get_url)
        self.result = response_API.json()
        print(self.result)
        container.remove(force=True, v=True)

    def predict(self, X_test: np.ndarray = None, y_test: np.ndarray = None):
        return self.result


if __name__ == "__main__":

    tmp = RunAutoSklearn()
    tmp.fit(None, None,  "auroc")



from AutoImblearn.components.api import BaseTransformerAPI
from imblearn.over_sampling import SMOTE, RandomOverSampler
from imblearn.under_sampling import RandomUnderSampler, OneSidedSelection, CondensedNearestNeighbour
from imblearn.combine import SMOTEENN
import numpy as np

import logging


rsps = {
    'rus': RandomUnderSampler(random_state=42),
    'ros': RandomOverSampler(random_state=42),
    'smote': SMOTE(random_state=42),
    # 'mwmote': MWMOTE(random_state=42),

    # 'combined': SMOTEENN(random_state=42, n_jobs=-1),
}

class RunImblearnSamplerAPI(BaseTransformerAPI):

    def get_hyperparameter_search_space(self):
        return {
        }

    def fit(self, args, X_train, y_train, X_test, y_test):
        resampler = rsps[self.params.model]

        params = self.dict_to_namespace()
        params.metric = args.metric


        logging.info("finished parameter setting")

        return

    def transform(self, X):
        return X
